# ----------------------------------------------------------------
# Import plugin that downloads market and ship vendor data from the
# Elite Dangerous mobile API.
# ----------------------------------------------------------------

import hashlib
import json
import pathlib
import random
import requests
import time
import base64
import webbrowser
import configparser

from datetime import datetime, timezone
from collections import namedtuple
from http import HTTPStatus
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import urlsplit, parse_qs

from .. import cache, csvexport, plugins, mapping, fs

import secrets


__version_info__ = (5, 0, 2)
__version__ = '.'.join(map(str, __version_info__))

# ----------------------------------------------------------------
# Deal with some differences in names between TD, ED and the API.
# ----------------------------------------------------------------

bracket_levels = ('?', 'L', 'M', 'H')

# Categories to ignore. Drones end up here. No idea what they are.
cat_ignore = [
    'NonMarketable',
]


class OAuthCallbackHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        split_url = urlsplit(self.path)
        if split_url.path == "/callback":
            parsed_url = parse_qs(split_url.query)
            self.server.callback_code = parsed_url.get("code", [None])[0]
            self.server.callback_state = parsed_url.get("state", [None])[0]
            self.send_response(HTTPStatus.OK)
            body_text = b"<p>You can close me now.</p>"
        else:
            self.send_response(HTTPStatus.NOT_IMPLEMENTED)
            body_text = b"<p>Something went wrong.</p>"
        self.end_headers()
        self.wfile.write(b"<html><head><title>EDAPI Frontier Login</title></head>")
        self.wfile.write(b"<body><h1>AUTHENTICATION</h1>")
        self.wfile.write(body_text)
        self.wfile.write(b"</body></html>")
    def log_message(self, format, *args):
        pass


class OAuthCallbackServer(object):
    def __init__(self, hostname, port, handler):
        myServer = HTTPServer
        myServer.callback_code = None
        myServer.callback_state = None
        self.httpd = myServer((hostname, port), handler)
        self.httpd.handle_request()
        self.httpd.server_close()


class EDAPI:
    '''
    A class that handles the Frontier ED API.
    '''

    _agent = "EDCD-TradeDangerousPluginEDAPI-%s" % __version__
    _basename = 'edapi'
    _configfile = _basename + '.config'

    def __init__(
        self,
        basename='edapi',
        debug=False,
        configfile=None,
        json_file=None,
        login=False
    ):
        '''
        Initialize
        '''

        # Build common file names from basename.
        self._basename = basename
        if configfile:
            self._configfile = configfile

        self.debug = debug
        self.login = login

        # If json_file was given, just load that instead.
        if json_file:
            with open(json_file) as file:
                self.profile = json.load(file)
                return

        # Setup the session.
        self.opener = requests.Session()

        # Setup config
        self.config = configparser.ConfigParser()
        self.config.read_dict({
            "frontier": {
                "AUTH_URL": "https://auth.frontierstore.net",
                "AUTH_URL_AUTH": "https://auth.frontierstore.net/auth",
                "AUTH_URL_TOKEN": "https://auth.frontierstore.net/token",
            },
            "companion": {
                "CAPI_LIVE_URL": "https://companion.orerve.net",
                "CAPI_BETA_URL": "https://pts-companion.orerve.net",
                "CLIENT_ID": "0d60c9fe-1ae3-4849-91e9-250db5de9d79",
                "REDIRECT_URI": "http://127.2.0.1:2989/callback",
            },
            "authorization": {}
        })
        self._authorization_set_config({})
        self.config.read(self._configfile)

        # If force login, kill the authorization
        if self.login:
            self._authorization_set_config({})

        # Grab the commander profile
        self.text = []
        self.profile = self.query_capi("/profile")
        market = self.query_capi("/market")
        shipyard = self.query_capi("/shipyard")

        # Grab the market, outfitting and shipyard data if needed
        portServices = self.profile['lastStarport'].get('services')
        if self.profile['commander']['docked'] and portServices:
            if portServices.get('commodities'):
                res = self.query_capi("/market")
                if int(res["id"]) == int(self.profile["lastStarport"]["id"]):
                    self.profile["lastStarport"].update(res)
            hasShipyard = portServices.get('shipyard')
            if hasShipyard or portServices.get('outfitting'):
                # the ships for the shipyard are not always returned the first time
                for attempt in range(3):
                    # try up to 3 times
                    res = self.query_capi("/shipyard")
                    if not hasShipyard or res.get('ships'):
                        break
                    if self.debug:
                        print("No shipyard in response, I'll try again in 5s")
                    time.sleep(5)
                if int(res["id"]) == int(self.profile["lastStarport"]["id"]):
                    self.profile["lastStarport"].update(res)

    def query_capi(self, capi_endpoint):
        self._authorization_check()
        response = self.opener.get(self.config["companion"]["CAPI_LIVE_URL"] + capi_endpoint)
        try:
            print(response.text)
            data = response.json()
            self.text.append(response.text)
        except:
            if self.debug:
                print('   URL:', response.url)
                print('status:', response.status_code)
                print('  text:', response.text)
                txtDebug = ""
            else:
                txtDebug = "\nTry with --debug and report this."
            raise plugins.PluginException(
                "Unable to parse JSON response for {}!"
                "\nTry to relogin with the 'login' option."
                "{}".format(capi_endpoint, txtDebug)
            )
        return data

    def _authorization_check(self):
        status_ok = True
        expires_at = self.config.getint("authorization", "expires_at")
        if self.debug:
            print("auth expires_at", time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(expires_at)))
        if (expires_at - time.time()) < 60:
            if self.debug:
                print("authorization expired")
            status_ok = False
            if self.config["authorization"]["refresh_token"]:
                status_ok = self._authorization_refresh()
            if not status_ok:
                status_ok = self._authorization_login()
            with open(self._configfile, "w") as c:
                self.config.write(c)

        if not status_ok:
            # Something terrible happend
            raise plugins.PluginException("Couldn't get frontier authorization.")

        # Setup session authorization
        self.opener.headers = {
            'User-Agent': self._agent,
            'Authorization': "%s %s" % (
                self.config['authorization']['token_type'],
                self.config['authorization']['access_token'],
            ),
        }

    def _authorization_set_config(self, auth_data):
        self.config.set("authorization", "access_token", auth_data.get("access_token", ""))
        self.config.set("authorization", "token_type", auth_data.get("token_type", ""))
        self.config.set("authorization", "expires_at", str(auth_data.get("expires_at", 0)))
        self.config.set("authorization", "refresh_token", auth_data.get("refresh_token", ""))

    def _authorization_token(self, data):
        expires_at = int(time.time())
        res = requests.post(self.config["frontier"]["AUTH_URL_TOKEN"], data=data)
        if self.debug:
            print(res, res.url)
            print(res.text)
        if res.status_code == requests.codes.ok:
            auth_data = res.json()
            auth_data['expires_at'] = expires_at + int(auth_data.get('expires_in', 0))
            self._authorization_set_config(auth_data)
            return True
        self._authorization_set_config({})
        return False

    def _authorization_refresh(self):
        data = {
            "grant_type": "refresh_token",
            "refresh_token": self.config["authorization"]["refresh_token"],
            "client_id": self.config["companion"]["CLIENT_ID"],
        }
        return self._authorization_token(data)

    def _authorization_login(self):
        session_state = secrets.token_urlsafe(36)
        code_verifier = secrets.token_urlsafe(36)
        code_digest = hashlib.sha256(code_verifier.encode()).digest()
        code_challenge = base64.urlsafe_b64encode(code_digest).decode().rstrip("=")
        data = {
            'response_type': 'code',
            'redirect_uri': self.config["companion"]["REDIRECT_URI"],
            'client_id': self.config["companion"]["CLIENT_ID"],
            'code_challenge': code_challenge,
            'code_challenge_method': 'S256',
            'state': session_state,
        }
        req = requests.Request("GET", self.config["frontier"]["AUTH_URL_AUTH"], params=data)
        pre = req.prepare()
        webbrowser.open_new_tab(pre.url)

        redirect_uri = urlsplit(self.config["companion"]["REDIRECT_URI"])
        oauth = OAuthCallbackServer(redirect_uri.hostname, redirect_uri.port, OAuthCallbackHandler)
        if oauth.httpd.callback_code and oauth.httpd.callback_state == session_state:
            data = {
                "grant_type": "authorization_code",
                "code": oauth.httpd.callback_code,
                "code_verifier": code_verifier,
                "client_id": self.config["companion"]["CLIENT_ID"],
                "redirect_uri": self.config["companion"]["REDIRECT_URI"],
            }
            return self._authorization_token(data)
        return False


class EDDN:
    _gateways = (
        'https://eddn.edcd.io:4430/upload/',
        # 'http://eddn-gateway.ed-td.space:8080/upload/',
    )

    _commodity_schemas = {
        'production': 'https://eddn.edcd.io/schemas/commodity/3',
        'test': 'https://eddn.edcd.io/schemas/commodity/3/test',
    }

    _shipyard_schemas = {
        'production': 'https://eddn.edcd.io/schemas/shipyard/2',
        'test': 'https://eddn.edcd.io/schemas/shipyard/2/test',
    }

    _outfitting_schemas = {
        'production': 'https://eddn.edcd.io/schemas/outfitting/2',
        'test': 'https://eddn.edcd.io/schemas/outfitting/2/test',
    }

    _debug = True

    # As of 1.3, ED reports four levels.
    _levels = (
        'Low',
        'Low',
        'Med',
        'High',
    )

    def __init__(
        self,
        uploaderID,
        noHash,
        softwareName,
        softwareVersion
    ):
        # Obfuscate uploaderID
        if noHash:
            self.uploaderID = uploaderID
        else:
            self.uploaderID = hashlib.sha1(uploaderID.encode('utf-8')).hexdigest()
        self.softwareName = softwareName
        self.softwareVersion = softwareVersion

    def postMessage(
        self,
        message,
        timestamp=0
    ):
        if timestamp:
            timestamp = datetime.fromtimestamp(timestamp).isoformat()
        else:
            timestamp = datetime.now(timezone.utc).astimezone().isoformat()

        message['message']['timestamp'] = timestamp

        url = random.choice(self._gateways)

        headers = {
            'content-type': 'application/json; charset=utf8'
        }

        if self._debug:
            print(
                json.dumps(
                    message,
                    sort_keys=True,
                    indent=4
                )
            )

        r = requests.post(
            url,
            headers=headers,
            data=json.dumps(
                message,
                ensure_ascii=False
            ).encode('utf8'),
            verify=True
        )

        r.raise_for_status()

    def publishCommodities(
        self,
        systemName,
        stationName,
        marketId,
        commodities,
        additional=None,
        timestamp=0
    ):
        message = {}

        message['$schemaRef'] = self._commodity_schemas[('test' if self._debug else 'production')]  # NOQA

        message['header'] = {
            'uploaderID': self.uploaderID,
            'softwareName': self.softwareName,
            'softwareVersion': self.softwareVersion
        }

        message['message'] = {
            'systemName': systemName,
            'stationName': stationName,
            'marketId': marketId,
            'commodities': commodities,
        }
        if additional:
            message['message'].update(additional)

        self.postMessage(message, timestamp)

    def publishShipyard(
        self,
        systemName,
        stationName,
        marketId,
        ships,
        timestamp=0
    ):
        message = {}

        message['$schemaRef'] = self._shipyard_schemas[('test' if self._debug else 'production')]  # NOQA

        message['header'] = {
            'uploaderID': self.uploaderID,
            'softwareName': self.softwareName,
            'softwareVersion': self.softwareVersion
        }

        message['message'] = {
            'systemName': systemName,
            'stationName': stationName,
            'marketId': marketId,
            'ships': ships,
        }

        self.postMessage(message, timestamp)

    def publishOutfitting(
        self,
        systemName,
        stationName,
        marketId,
        modules,
        timestamp=0
    ):
        message = {}

        message['$schemaRef'] = self._outfitting_schemas[('test' if self._debug else 'production')]  # NOQA

        message['header'] = {
            'uploaderID': self.uploaderID,
            'softwareName': self.softwareName,
            'softwareVersion': self.softwareVersion
        }

        message['message'] = {
            'systemName': systemName,
            'stationName': stationName,
            'marketId': marketId,
            'modules': modules,
        }

        self.postMessage(message, timestamp)


class ImportPlugin(plugins.ImportPluginBase):
    """
    Plugin that downloads market and ship vendor data from the Elite Dangerous
    mobile API.
    """

    pluginOptions = {
        'csvs': 'Merge shipyards into ShipVendor.csv.',
        'edcd': 'Call the EDCD plugin first.',
        'eddn': 'Post market, shipyard and outfitting to EDDN.',
        'name': 'Do not obfuscate commander name for EDDN submit.',
        'save': 'Save the API response (tmp/profile.YYYYMMDD_HHMMSS.json).',
        'tdh': 'Save the API response for TDH (tmp/tdh_profile.json).',
        'test': 'Test the plugin with a json file (test=[FILENAME]).',
        'warn': 'Ask for station update if a API<->DB diff is encountered.',
        'login': 'Ask for login credentials.',
    }

    configFile = "edapi.config"

    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)

        self.filename = self.defaultImportFile
        self.configPath = tdb.dataPath / pathlib.Path(ImportPlugin.configFile)

    def askForStationData(self, system, stnName=None, station=None):
        """
        Ask for new or updated station data
        """
        tdb, tdenv = self.tdb, self.tdenv
        askForData = False

        stnDefault = namedtuple(
            'stnDefault', [
                'lsFromStar','market','blackMarket','shipyard','maxPadSize',
                'outfitting','rearm','refuel','repair','planetary',
            ]
        )

        def tellUserAPIResponse(defName, defValue):
            if defValue == "Y":
                tdenv.NOTE("{:>12} in API response", defName)
            else:
                tdenv.NOTE("{:>12} NOT in API response", defName)

        def getYNfromObject(obj, key, val=None):
            if val:
                return "Y" if obj.get(key) == val else "N"
            else:
                return "Y" if key in obj else "N"

        # defaults from API response are not reliable!
        checkStarport = self.edAPI.profile['lastStarport']
        defMarket     = getYNfromObject(checkStarport, 'commodities')
        defShipyard   = getYNfromObject(checkStarport, 'ships')
        defOutfitting = getYNfromObject(checkStarport, 'modules')
        tellUserAPIResponse("'Outfitting'", defOutfitting)
        tellUserAPIResponse("'ShipYard'", defShipyard)
        tellUserAPIResponse("'Market'", defMarket)

        def warnAPIResponse(checkName, checkYN):
            # no warning if unknown
            if checkYN == "?": return False
            warnText = (
                "The station should{s} have a {what}, "
                "but the API did{d} return one."
            )
            if checkYN == "Y":
                s, d = "", "n't"
            else:
                s, d = "n't", ""

            tdenv.WARN(warnText, what=checkName, s=s, d=d)
            return True if self.getOption('warn') else False

        # station services since ED update 2.4
        checkServices = checkStarport.get('services', None)
        if checkServices:
            if station:
                tdenv.NOTE('Station known.')
                stnlsFromStar = station.lsFromStar
                stnmaxPadSize = station.maxPadSize
                stnplanetary  = station.planetary
            else:
                tdenv.NOTE('Station unknown.')
                stnlsFromStar = 0
                stnmaxPadSize = "?"
                stnplanetary  = "?"
            tdenv.NOTE("Found station services.")
            if checkStarport.get('outpostType', None) == 'starport':
                # only the big one can be detected
                stnmaxPadSize = "L"
                stnplanetary  = "N"
            defStation = stnDefault(
                lsFromStar = stnlsFromStar,
                market = getYNfromObject(checkServices, 'commodities', val='ok'),
                blackMarket = getYNfromObject(checkServices, 'blackmarket', val='ok'),
                shipyard = getYNfromObject(checkServices, 'shipyard', val='ok'),
                maxPadSize = stnmaxPadSize,
                outfitting = getYNfromObject(checkServices, 'outfitting', val='ok'),
                rearm = getYNfromObject(checkServices, 'rearm', val='ok'),
                refuel = getYNfromObject(checkServices, 'refuel', val='ok'),
                repair = getYNfromObject(checkServices, 'repair', val='ok'),
                planetary = stnplanetary,
            )
        elif station:
            tdenv.NOTE('Station known.')
            defStation = stnDefault(
                lsFromStar = station.lsFromStar,
                market = defMarket if station.market == "?" else station.market,
                blackMarket = station.blackMarket,
                shipyard = defShipyard if station.shipyard == "?" else station.shipyard,
                maxPadSize = station.maxPadSize,
                outfitting = defOutfitting if station.outfitting == "?" else station.outfitting,
                rearm = station.rearm,
                refuel = station.refuel,
                repair = station.repair,
                planetary = station.planetary,
            )
        else:
            tdenv.NOTE('Station unknown.')
            defStation = stnDefault(
                lsFromStar  = 0,   market     = defMarket,
                blackMarket = "?", shipyard   = defShipyard,
                maxPadSize  = "?", outfitting = defOutfitting,
                rearm       = "?", refuel     = "?",
                repair      = "?", planetary  = "?",
            )

        warning = False
        if defStation.outfitting != defOutfitting:
            warning |= warnAPIResponse('outfitting', defStation.outfitting)
        if defStation.shipyard != defShipyard:
            warning |= warnAPIResponse('shipyard', defStation.shipyard)
        if defStation.market != defMarket:
            warning |= warnAPIResponse('market', defStation.market)
        if warning:
            tdenv.WARN("Please update station data with correct values.")
            tdenv.WARN("(Fields will be marked with an leading asterisk '*')")
            askForData = True
        if ((defStation.lsFromStar == 0) or ("?" in defStation)):
           askForData = True

        newStation = {}
        for key in defStation._fields:
            newStation[key] = getattr(defStation, key)

        if askForData:
            tdenv.NOTE("Values in brackets are the default.")
            lsFromStar = input(
                " Stn/Ls..............[{}]: ".format(defStation.lsFromStar)
            ) or defStation.lsFromStar
            try:
                lsFromStar = int(float(lsFromStar)+0.5)
            except:
                print("That doesn't seem to be a number. Defaulting to zero.")
                lsFromStar = defStation.lsFromStar
            newStation['lsFromStar'] = lsFromStar

            for askText, askField, markValue in [
                ('Pad Size....(s,m,l) ', 'maxPadSize',  defStation.maxPadSize),
                ('Planetary.....(y,n) ', 'planetary',   defStation.planetary),
                ('B/Market......(y,n) ', 'blackMarket', defStation.blackMarket),
                ('Refuel........(y,n) ', 'refuel',      defStation.refuel),
                ('Repair........(y,n) ', 'repair',      defStation.repair),
                ('Restock.......(y,n) ', 'rearm',       defStation.rearm),
                ('Outfitting....(y,n) ', 'outfitting',  defOutfitting),
                ('Shipyard......(y,n) ', 'shipyard',    defShipyard),
                ('Market........(y,n) ', 'market',      defMarket),
            ]:
                defValue = getattr(defStation, askField)
                if defValue != markValue:
                    mark = "*"
                else:
                    mark = " "
                askValue = input(
                    "{}{}[{}]: ".format(mark, askText, defValue)
                ) or defValue
                newStation[askField] = askValue

        else:
            def _detail(value, source):
                detail = source[value]
                if detail == '?':
                    detail += ' [unknown]'
                return detail

            ls = newStation['lsFromStar']
            print(" Stn/Ls....:", ls, '[unknown]' if ls == 0 else '')
            print(" Pad Size..:", _detail(newStation['maxPadSize'], tdb.padSizes))
            print(" Planetary.:", _detail(newStation['planetary'], tdb.planetStates))
            print(" B/Market..:", _detail(newStation['blackMarket'], tdb.marketStates))
            print(" Refuel....:", _detail(newStation['refuel'], tdb.marketStates))
            print(" Repair....:", _detail(newStation['repair'], tdb.marketStates))
            print(" Restock...:", _detail(newStation['rearm'], tdb.marketStates))
            print(" Outfitting:", _detail(newStation['outfitting'], tdb.marketStates))
            print(" Shipyard..:", _detail(newStation['shipyard'], tdb.marketStates))
            print(" Market....:", _detail(newStation['market'], tdb.marketStates))

        exportCSV = False
        if not station:
            station = tdb.addLocalStation(
                system=system,
                name=stnName,
                lsFromStar=newStation['lsFromStar'],
                blackMarket=newStation['blackMarket'],
                maxPadSize=newStation['maxPadSize'],
                market=newStation['market'],
                shipyard=newStation['shipyard'],
                outfitting=newStation['outfitting'],
                rearm=newStation['rearm'],
                refuel=newStation['refuel'],
                repair=newStation['repair'],
                planetary=newStation['planetary'],
            )
            exportCSV = True
        else:
            # let the function check for changes
            if tdb.updateLocalStation(
                station=station,
                lsFromStar=newStation['lsFromStar'],
                blackMarket=newStation['blackMarket'],
                maxPadSize=newStation['maxPadSize'],
                market=newStation['market'],
                shipyard=newStation['shipyard'],
                outfitting=newStation['outfitting'],
                rearm=newStation['rearm'],
                refuel=newStation['refuel'],
                repair=newStation['repair'],
                planetary=newStation['planetary'],
            ):
                exportCSV = True

        if exportCSV:
            lines, csvPath = csvexport.exportTableToFile(
                tdb,
                tdenv,
                "Station",
            )
            tdenv.DEBUG0("{} updated.", csvPath)
        return station

    def run(self):
        tdb, tdenv = self.tdb, self.tdenv

        # first check for EDCD
        if self.getOption("edcd"):
            # Call the EDCD plugin
            try:
                import plugins.edcd_plug as EDCD
            except:
                raise plugins.PluginException("EDCD plugin not found.")
            tdenv.NOTE("Calling the EDCD plugin.")
            edcdPlugin = EDCD.ImportPlugin(tdb, tdenv)
            edcdPlugin.options["csvs"] = True
            edcdPlugin.run()
            tdenv.NOTE("Going back to EDAPI.\n")

        # now load the mapping tables
        itemMap = mapping.FDEVMappingItems(tdb, tdenv)
        shipMap = mapping.FDEVMappingShips(tdb, tdenv)

        # Connect to the API, authenticate, and pull down the commander
        # /profile.
        if self.getOption("test"):
            tdenv.WARN("#############################")
            tdenv.WARN("###  EDAPI in test mode.  ###")
            tdenv.WARN("#############################")
            apiED = namedtuple('EDAPI', ['profile','text'])
            try:
                proPath = pathlib.Path(self.getOption("test"))
            except TypeError:
                raise plugins.PluginException(
                    "Option 'test' must be a file name"
                )
            if proPath.exists():
                with proPath.open() as proFile:
                    proData = json.load(proFile)
                    if isinstance(proData, list):
                        # since 4.3.0: list(profile, market, shipyard)
                        testProfile = proData[0]
                        for data in proData[1:]:
                            if int(data["id"]) == int(testProfile["lastStarport"]["id"]):
                                testProfile["lastStarport"].update(data)
                    else:
                        testProfile = proData
                    api = apiED(
                        profile = testProfile,
                        text = '{{"mode":"test","file":"{}"}}'.format(str(proPath))
                    )
            else:
                raise plugins.PluginException(
                    "JSON-file '{}' not found.".format(str(proPath))
                )
        else:
            api = EDAPI(
                configfile=str(self.configPath),
                login=self.getOption('login'),
                debug=tdenv.debug,
            )
        self.edAPI = api
        
        fs.ensurefolder(tdenv.tmpDir)
        tdh_path = 'tdh_profile.json'
        if self.getOption("tdh"):
            self.options["save"] = True
            if tdh_path.exists():
                tdh_path.unlink()

        # save profile if requested
        if self.getOption("save"):
            saveName = tdh_path if self.getOption("tdh") else 'profile.' + time.strftime('%Y%m%d_%H%M%S') + '.json'
            savePath = tdenv.tmpDir / pathlib.Path(saveName)
            with open(savePath, 'w', encoding="utf-8") as saveFile:
                if isinstance(api.text, list):
                    # since 4.3.0: list(profile, market, shipyard)
                    tdenv.DEBUG0("{}",api.text)
                    saveFile.write('{{"profile":{}}}'.format(api.text[0]))
                else:
                    saveFile.write(api.text)
                print('API response saved to: {}'.format(savePath))

        # If TDH is calling the plugin, nothing else needs to be done
        # now that the file has been created.
        if self.getOption("tdh"):
            return False

        # Sanity check that the commander is docked. Otherwise we will get a
        # mismatch between the last system and last station.
        if not api.profile['commander']['docked']:
            print('Commander not docked. Aborting!')
            return False

        # Figure out where we are.
        sysName = api.profile['lastSystem']['name']
        stnName = api.profile['lastStarport']['name']
        marketId = int(api.profile['lastStarport']['id'])
        print('@{}/{} (ID: {})'.format(sysName.upper(), stnName, marketId))

        # Reload the cache.
        tdenv.DEBUG0("Checking the cache")
        tdb.close()
        tdb.reloadCache()
        tdb.load(
            maxSystemLinkLy=tdenv.maxSystemLinkLy,
        )
        tdb.close()

        # Check to see if this system is in the database
        try:
            system = tdb.lookupSystem(sysName)
        except LookupError:
            raise plugins.PluginException(
                "System '{}' unknown.".format(sysName)
            )

        # Check to see if this station is in the database
        try:
            station = tdb.lookupStation(stnName, system)
        except LookupError:
            station = None

        # New or update station data
        station = self.askForStationData(system, stnName=stnName, station=station)

        # If a shipyard exists, make the ship lists
        shipCost = {}
        shipList = []
        eddn_ships = []
        if ((station.shipyard == "Y") and
            ('ships' in api.profile['lastStarport'])
        ):
            if 'shipyard_list' in api.profile['lastStarport']['ships']:
                if len(api.profile['lastStarport']['ships']['shipyard_list']):
                    for ship in api.profile['lastStarport']['ships']['shipyard_list'].values():
                        shipName = shipMap.mapID(ship['id'], ship['name'])
                        shipCost[shipName] = ship['basevalue']
                        shipList.append(shipName)
                        eddn_ships.append(ship['name'])

            if 'unavailable_list' in api.profile['lastStarport']['ships']:
                for ship in api.profile['lastStarport']['ships']['unavailable_list']:
                    shipName = shipMap.mapID(ship['id'], ship['name'])
                    shipCost[shipName] = ship['basevalue']
                    shipList.append(shipName)
                    eddn_ships.append(ship['name'])

        if self.getOption("csvs"):
            addShipList = set()
            delShipList = set()
            addRows = delRows = 0
            db = tdb.getDB()
            if station.shipyard == "N":
                # delete all ships if there is no shipyard
                delRows = db.execute(
                    """
                    DELETE FROM ShipVendor
                     WHERE station_id = ?
                    """,
                    [station.ID]
                ).rowcount

            if len(shipList):
                # and now update the shipyard list
                # we go through all ships to decide if a ship needs to be
                # added or deleted from the shipyard
                for shipID in tdb.shipByID:
                    shipName = tdb.shipByID[shipID].dbname
                    if shipName in shipList:
                        # check for ship discount, costTD = 100%
                        # python builtin round() uses "Round half to even"
                        # but we need commercial rounding, so we do it ourself
                        costTD = tdb.shipByID[shipID].cost
                        costED = int((shipCost[shipName]+5)/10)*10
                        if costTD != costED:
                            prozED = int(shipCost[shipName]*100/costTD+0.5)-100
                            tdenv.NOTE(
                                "CostDiff {}: {} != {} ({}%)",
                                shipName, costTD, costED, prozED
                            )
                        # add the ship to the shipyard
                        shipSQL = (
                            "INSERT OR IGNORE"
                             " INTO ShipVendor(station_id, ship_id)"
                           " VALUES(?, ?)"
                        )
                        tdenv.DEBUG0(shipSQL.replace("?", "{}"), station.ID, shipID)
                        rc = db.execute(shipSQL, [station.ID, shipID]).rowcount
                        if rc:
                            addRows += rc
                            addShipList.add(shipName)
                        # remove ship from the list
                        shipList.remove(shipName)
                    else:
                        # delete the ship from the shipyard
                        shipSQL = (
                            "DELETE FROM ShipVendor"
                            " WHERE station_id = ?"
                              " AND ship_id = ?"
                        )
                        tdenv.DEBUG0(shipSQL.replace("?", "{}"), station.ID, shipID)
                        rc = db.execute(shipSQL, [station.ID, shipID]).rowcount
                        if rc:
                            delRows += rc
                            delShipList.add(shipName)

                if len(shipList):
                    tdenv.WARN("unknown Ship(s): {}", ",".join(shipList))

            db.commit()
            if (addRows + delRows) > 0:
                if addRows > 0:
                    tdenv.NOTE(
                        "Added {} ({}) ships in '{}' shipyard.",
                        addRows, ", ".join(sorted(addShipList)), station.name()
                    )
                if delRows > 0:
                    tdenv.NOTE(
                        "Deleted {} ({}) ships in '{}' shipyard.",
                        delRows, ", ".join(sorted(delShipList)), station.name()
                    )
                lines, csvPath = csvexport.exportTableToFile(
                    tdb,
                    tdenv,
                    "ShipVendor",
                )
                tdenv.DEBUG0("{} updated.", csvPath)

        # If a market exists, make the item lists
        itemList = []
        eddn_market = []
        if ((station.market == "Y") and
            ('commodities' in api.profile['lastStarport'])
        ):
            for commodity in api.profile['lastStarport']['commodities']:
                if commodity['categoryname'] in cat_ignore:
                    continue

                if commodity.get('legality', '') != '':
                    # ignore if present and not empty
                    continue

                locName = commodity.get('locName', commodity['name'])
                itmName = itemMap.mapID(commodity['id'], locName)

                def commodity_int(key):
                    try:
                        ret = int(float(commodity[key])+0.5)
                    except (ValueError, KeyError):
                        ret = 0
                    return ret

                itmSupply      = commodity_int('stock')
                itmDemand      = commodity_int('demand')
                itmSupplyLevel = commodity_int('stockBracket')
                itmDemandLevel = commodity_int('demandBracket')
                itmBuyPrice    = commodity_int('buyPrice')
                itmSellPrice   = commodity_int('sellPrice')

                if itmSupplyLevel == 0 or itmBuyPrice == 0:
                    # If there is not stockBracket or buyPrice, ignore stock
                    itmBuyPrice = 0
                    itmSupply = 0
                    itmSupplyLevel = 0
                    tdSupply = "-"
                    tdDemand = "{}{}".format(
                        itmDemand,
                        bracket_levels[itmDemandLevel]
                    )
                else:
                    # otherwise don't care about demand
                    itmDemand = 0
                    itmDemandLevel = 0
                    tdDemand = "?"
                    tdSupply = "{}{}".format(
                        itmSupply,
                        bracket_levels[itmSupplyLevel]
                    )

                # ignore items without supply or demand bracket (TD only)
                if itmSupplyLevel > 0 or itmDemandLevel > 0:
                    itemTD = (
                        itmName,
                        itmSellPrice, itmBuyPrice,
                        tdDemand, tdSupply,
                    )
                    itemList.append(itemTD)

                # Populate EDDN
                if self.getOption("eddn"):
                    itemEDDN = {
                        "name":          commodity['name'],
                        "meanPrice":     commodity_int('meanPrice'),
                        "buyPrice":      commodity_int('buyPrice'),
                        "stock":         commodity_int('stock'),
                        "stockBracket":  commodity['stockBracket'],
                        "sellPrice":     commodity_int('sellPrice'),
                        "demand":        commodity_int('demand'),
                        "demandBracket": commodity['demandBracket'],
                    }
                    if len(commodity['statusFlags']) > 0:
                        itemEDDN["statusFlags"] = commodity['statusFlags']
                    eddn_market.append(itemEDDN)

        if itemList:
            # Create the import file.
            with open(self.filename, 'w', encoding="utf-8") as f:
                # write System/Station line
                f.write("@ {}/{}\n".format(sysName, stnName))

                # write Item lines (category lines are not needed)
                for itemTD in itemList:
                    f.write("\t\t%s %s %s %s %s\n" % itemTD)

            tdenv.ignoreUnknown = True
            cache.importDataFromFile(
                tdb,
                tdenv,
                pathlib.Path(self.filename),
            )

        # Import EDDN
        if self.getOption("eddn"):
            con = EDDN(
                api.profile['commander']['name'],
                self.getOption("name"),
                'EDAPI Trade Dangerous Plugin',
                __version__
            )
            if self.getOption("test"):
                con._debug = True
            else:
                con._debug = False

            if eddn_market:
                print('Posting commodities to EDDN...')
                eddn_additional = {}
                if 'economies' in api.profile['lastStarport']:
                    eddn_additional['economies'] = []
                    for economy in api.profile['lastStarport']['economies'].values():
                        eddn_additional['economies'].append(economy)
                if 'prohibited' in api.profile['lastStarport']:
                    eddn_additional['prohibited'] = []
                    for item in api.profile['lastStarport']['prohibited'].values():
                        eddn_additional['prohibited'].append(item)

                con.publishCommodities(
                    sysName,
                    stnName,
                    marketId,
                    eddn_market,
                    additional=eddn_additional
                )

            if eddn_ships:
                print('Posting shipyard to EDDN...')
                con.publishShipyard(
                    sysName,
                    stnName,
                    marketId,
                    eddn_ships
                )

            if ((station.outfitting == "Y") and
                ('modules' in api.profile['lastStarport'] and
                len(api.profile['lastStarport']['modules']))
            ):
                eddn_modules = []
                for module in api.profile['lastStarport']['modules'].values():
                    # see: https://github.com/EDSM-NET/EDDN/wiki
                    addModule = False
                    if module['name'].startswith(('Hpt_', 'Int_')) or module['name'].find('_Armour_') > 0:
                        if module.get('sku', None) in (
                            None, 'ELITE_HORIZONS_V_PLANETARY_LANDINGS'
                        ):
                            if module['name'] != 'Int_PlanetApproachSuite':
                                addModule = True
                    if addModule:
                        eddn_modules.append(module['name'])
                    elif self.getOption("test"):
                        tdenv.NOTE("Ignored module ID: {}, name: {}", module['id'], module['name'])
                if eddn_modules:
                    print('Posting outfitting to EDDN...')
                    con.publishOutfitting(
                        sysName,
                        stnName,
                        marketId,
                        sorted(eddn_modules)
                    )

        # We did all the work
        return False


"""
x52pro.py - DirectOutputDevice for SaitekX52Pro

Version: 0.2
Author: Frazzle
Version: 0.3
Author: Oliver "kfsone" Smith <oliver@kfs.org> 2014

Description: Python class implementing specific functions for Saitek X52 Pro


TODO:
    * Error handling and exceptions
"""

from __future__ import absolute_import, with_statement, print_function, division, unicode_literals

from mfd.saitek.directoutput import DirectOutputDevice


class X52Pro(DirectOutputDevice):
    class Page(object):
        _lines = [ str(), str(), str() ]
        _leds  = dict()

        def __init__(self, device, page_id, name, active):
            self.device = device
            self.page_id = page_id
            self.name = name
            self.device.AddPage(self.page_id, name, 1 if active else 0)
            self.active = active

        def __del__(self, *args, **kwargs):
            try:
                self.device.RemovePage(self.page_id)
            except AttributeError:
                pass

        def __getitem__(self, key):
            return self._lines[key]

        def __setitem__(self, key, value):
            self._lines[key] = value
            if self.active:
                self.device.SetString(self.page_id, key, value)

        def activate(self):
            if self.active == True:
                return
            self.device.AddPage(self.page_id, self.name, 1)

        def refresh(self):
            # Resend strings to the display
            for lineNo, string in enumerate(self._lines):
                self.device.SetString(self.page_id, lineNo, string)
            for led, value in self._leds:
                self.device.SetLed(self.page_id, led, 1 if value else 0)

        def set_led(self, led, value):
            self._leds[led] = value
            if self.active:
                self.device.SetLed(self.page_id, led, 1 if value else 0)

        def set_led_colour(self, value, led_red, led_green):
            if value == "red":
                self.set_led(led_red, 1)
                self.set_led(led_green, 0)
            elif value == "green":
                self.set_led(led_red, 0)
                self.set_led(led_green, 1)
            elif value == "orange":
                self.set_led(led_red, 1)
                self.set_led(led_green, 1)
            elif value == "off":
                self.set_led(led_red, 0)
                self.set_led(led_green, 0)

        def fire(self, value):
            self.set_led(0, value)

        def fire_a(self, value):
            self.set_led_colour(value, 1, 2)

        def fire_b(self, value):
            self.set_led_colour(value, 3, 4)

        def fire_d(self, value):
            self.set_led_colour(value, 5, 6)

        def fire_e(self, value):
            self.set_led_colour(value, 7, 8)

        def toggle_1_2(self, value):
            self.set_led_colour(value, 9, 10)

        def toggle_3_4(self, value):
            self.set_led_colour(value, 11, 12)

        def toggle_5_6(self, value):
            self.set_led_colour(value, 13, 14)

        def pov_2(self, value):
            self.set_led_colour(value, 15, 16)

        def clutch(self, value):
            self.set_led_colour(value, 17, 18)

        def throttle_axis(self, value):
            self.set_led(19, value)

    pages = {}
    _page_counter = 0

    def add_page(self, name, active=True):
        page = self.pages[name] = self.Page(self, self._page_counter, name, active=active)
        self._page_counter += 1
        return page

    def remove_page(self, name):
        del self.pages[name]


    def OnPage(self, page_id, activated):
        super().OnPage(page_id, activated)
        for page in self.pages.values():
            if page.page_id == page_id:
                print("Found the page", page_id, activated)
                if activated:
                    page.refresh()
                else:
                    page.active = False
                return


    def OnSoftButton(self, *args, **kwargs):
        super().OnSoftButton(*args, **kwargs)
        print("*** ON SOFT BUTTON")


    def finish(self):
        for page in self.pages:
            del page
        super().finish()


if __name__ == '__main__':
    x52 = X52Pro(debug_level=1)
    print("X52 Connected")

    x52.add_page("Page1")
    print("Page1 page added")

    x52.pages["Page1"][0] = "Test String"
    try:
        x52.pages["Page1"][5] = "FAIL"
        raise ValueError("Your stick thinks it has 5 lines of MFD, I think it lies.")
    except IndexError:
        print("Your MFD has the correct number of lines of text")

    x52.pages["Page1"][1] = "-- Page 1 [0]"

    fireToggle = False
    x52.pages["Page1"].fire(fireToggle)
    x52.pages["Page1"].fire_a("amber")
    x52.pages["Page1"].toggle_3_4("red")
    x52.pages["Page1"].throttle_axis(~fireToggle)

    print("Adding second page")
    x52.add_page("Page2", active=False)
    x52.pages["Page2"][0] = "Second Page Is Right Here"
    x52.pages["Page2"][1] = "-- Page 2 [1]"

    print("Looping")

    import time, sys

    loopNo = 0
    while True:
        try:
            loopNo += 1
            x52.pages["Page1"][2] = "Loop #" + str(loopNo)
            time.sleep(0.25)
            x52.pages["Page1"].fire_a("red")
            time.sleep(0.25)
            x52.pages["Page1"].fire_a("orange")
            time.sleep(0.25)
            x52.pages["Page1"].fire_a("green")
            time.sleep(0.25)
            x52.pages["Page1"].fire_a("off")
            fireToggle = ~fireToggle
            x52.pages["Page1"].fire(fireToggle)
            x52.pages["Page1"].throttle_axis(~fireToggle)
        except Exception as e:
            print(e)
            x52.finish()
            sys.exit()




# -*- coding: utf-8 -*-
import os
import re
import pathlib
import time as _time
from datetime import datetime, time, timedelta, timezone

from .. import tradedb, tradeenv, csvexport
from . import PluginException, ImportPluginBase

def snapToGrid32(val):
    try:
        val = float(val)
        if val < 0: corr = -0.5
        else:       corr = +0.5
        pos = int(val*32+corr)/32
    except:
        pos = None
        pass
    return pos


class ImportPlugin(ImportPluginBase):
    """
    Plugin that parses the netLog file and add or update the system.
    """

    logGlob = "netLog.*.log"
    ADDED_NAME = 'netLog'
    LOGDIR_NAME = "FDEVLOGDIR"
    DATE_FORMATS = {
         2: ("%y",       "YY",         "%y"),
         4: ("%Y",       "YYYY",       "%y"),
         5: ("%y-%m",    "YY-MM",      "%y%m"),
         7: ("%Y-%m",    "YYYY-MM",    "%y%m"),
         8: ("%y-%m-%d", "YY-MM-DD",   "%y%m%d"),
        10: ("%Y-%m-%d", "YYYY-MM-DD", "%y%m%d"),
    }
    ignoreSysNames  = [
        'TRAINING',
        'DESTINATION',
    ]

    pluginOptions = {
        'show': "Only show the system name and position. Don't update the DB.",
        'last': "Only parse the last (newest) netLog file.",
        'date': "Only parse netLog files from date, format=[YY]YY[-MM[-DD]].",
    }

    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)

        logDirName = os.getenv(self.LOGDIR_NAME, None)
        if not logDirName:
            raise PluginException(
                "Environment variable '{}' not set "
                "(see 'README.md' for help)."
                .format(self.LOGDIR_NAME)
            )
        tdenv.NOTE("{}={}", self.LOGDIR_NAME, logDirName)

        self.logPath = pathlib.Path(logDirName)
        if not self.logPath.is_dir():
            raise PluginException(
                "{}: is not a directory.".format(
                    str(self.logPath)
                )
            )

    def parseLogDirList(self):
        """
        parse netLog files
        """
        # HEADER: 16-07-02-00:18 Mitteleurop�ische Sommerzeit  (22:18 GMT) - part 1
        # SYSTEM: {00:20:24} System:"Caelinus" StarPos:(0.188,-18.625,52.063)ly  NormalFlight
        # or (since 2.3)
        # HEADER: ============================================
        # HEADER: Logs/netLog.170430120645.01.log (part 1)
        # HEADER: 2017-04-30 12:06 Mitteleurop�ische Sommerzeit
        # HEADER: ============================================
        # SYSTEM: {10:13:33GMT 407.863s} System:"Huokang" StarPos:(-12.188,35.469,-25.281)ly  NormalFlight
        tdb, tdenv = self.tdb, self.tdenv
        optShow = self.getOption("show")

        oldHeadRegEx = re.compile("^(?P<headDateTime>\d\d-\d\d-\d\d-\d\d:\d\d)\s+(?P<headTZName>.*[^\s])\s+(?P<headTimeGMT>\(.*GMT\))")
        newHeadRegEx = re.compile("^(?P<headDateTime>\d\d\d\d-\d\d-\d\d\s+\d\d:\d\d)\s+(?P<headTZName>.*[^\s])")

        sysRegEx  = re.compile('^\{[^\}]+\}\s+System:"(?P<sysName>[^"]+)".*StarPos:\((?P<sysPos>[^)]+)\)ly')
        dateRegEx = re.compile('^\{(?P<logTime>\d\d:\d\d:\d\d)')

        def calcSeconds(h=0, m=0, s=0):
            return 3600*h + 60*m + s

        sysCount = 0
        logSysList = {}
        for filePath in self.filePathList:
            tdenv.NOTE("parsing '{}'", filePath.name)
            oldCount = sysCount
            with filePath.open() as logFile:
                headDate, headMatch = None, None
                lineCount = 0
                statHeader = True
                for line in logFile:
                    lineCount += 1
                    line = line.strip('\r\n')
                    if statHeader:
                        # parse header line to get the date and timezone
                        tdenv.DEBUG0(" HEADER: {}", line.replace("{", "{{").replace("}", "}}"))
                        if lineCount == 1:
                            # old format
                            headMatch = oldHeadRegEx.match(line)
                            timeFormat = '%y-%m-%d-%H:%M'
                        if lineCount == 3:
                            # new format since 2.3
                            headMatch = newHeadRegEx.match(line)
                            timeFormat = '%Y-%m-%d %H:%M'
                        if headMatch:
                            headDate = headMatch.group('headDateTime')
                            headTZName = headMatch.group('headTZName')
                            if headTZName == _time.tzname[1]:
                                # daylight saving time
                                headTZInfo = timedelta(seconds = -_time.altzone)
                            else:
                                # normal time
                                headTZInfo = timedelta(seconds = -_time.timezone)
                            tdenv.DEBUG0(" HEADER: Date {}".format(headDate))
                            tdenv.DEBUG0(" HEADER: TZInfo {}".format(headTZInfo))
                            try:
                                # convert it into something useable
                                headDate = datetime.fromtimestamp(
                                    _time.mktime(
                                        _time.strptime(headDate, timeFormat)
                                    ),
                                    timezone(headTZInfo)
                                )
                            except:
                                headDate = None
                                pass
                        if not headDate:
                            if lineCount > 3:
                               raise PluginException("Doesn't seem do be a FDEV netLog file")
                        else:
                            statHeader = False
                            if lineCount == 3:
                                # new format since 2.3, switch to UTC
                                headDate = headDate.astimezone()
                            tdenv.DEBUG0("   DATE: {}", headDate)
                            headSecs = calcSeconds(headDate.hour, headDate.minute, headDate.second)
                            lastDate = logDate = headDate
                            lastSecs = logSecs = headSecs
                    else:
                        tdenv.DEBUG1("LOGLINE: {}", line.replace("{", "{{").replace("}", "}}"))
                        # check every line for new time to enhance the lastDate
                        # use time difference because of different timezone usage
                        logTimeMatch = dateRegEx.match(line)
                        if logTimeMatch:
                            h, m, s = logTimeMatch.group('logTime').split(":")
                            logSecs = calcSeconds(int(h), int(m), int(s))
                            logDiff = logSecs - lastSecs
                            if logDiff < 0:
                                # it's a new day
                                logDiff += 86400
                            logDate = lastDate + timedelta(seconds=logDiff)
                            tdenv.DEBUG1("LOGDATE: {}", logDate)

                        sysMatch = sysRegEx.match(line)
                        if sysMatch:
                            # we found a system, yeah
                            sysDate = logDate
                            sysName = sysMatch.group('sysName')
                            sysPos  = sysMatch.group('sysPos')
                            sysPosX, sysPosY, sysPosZ = sysPos.split(',')
                            sysPosX = snapToGrid32(sysPosX)
                            sysPosY = snapToGrid32(sysPosY)
                            sysPosZ = snapToGrid32(sysPosZ)
                            tdenv.DEBUG0(" SYSTEM: {} {} {} {} {}", sysDate, sysName, sysPosX, sysPosY, sysPosZ)
                            logSysList[sysName] = (sysPosX, sysPosY, sysPosZ, sysDate)

                        lastDate = logDate
                        lastSecs = logSecs
            sysCount = len(logSysList)
            tdenv.NOTE("Found {} System(s).", sysCount-oldCount)

        if not optShow:
            try:
                idNetLog = tdb.lookupAdded(self.ADDED_NAME)
            except KeyError:
                tdenv.WARN("Entry '{}' not found in 'Added' table.", self.ADDED_NAME)
                tdenv.WARN("Trying to add it myself.")
                db = tdb.getDB()
                cur = db.cursor()
                cur.execute(
                    """
                    INSERT INTO Added(name) VALUES(?)
                    """,
                    [self.ADDED_NAME]
                )
                db.commit()
                tdenv.NOTE("Export Table 'Added'")
                _, path = csvexport.exportTableToFile(tdb, tdenv, "Added")
                pass

        addCount = 0
        oldCount = 0
        newCount = 0
        for sysName in logSysList:
            sysPosX, sysPosY, sysPosZ, sysDate = logSysList[sysName]
            utcDate = sysDate.astimezone(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
            tdenv.DEBUG0("log system '{}' ({}, {}, {}, '{}')",
                sysName, sysPosX, sysPosY, sysPosZ, utcDate
            )
            if sysName.upper() in self.ignoreSysNames:
                tdenv.NOTE("Ignoring system: '{}'", sysName)
                continue
            systemTD = tdb.systemByName.get(sysName.upper(), None)
            if systemTD:
                # we allready know the system, check coords
                tdenv.DEBUG0("Old system '{}' ({}, {}, {})",
                    systemTD.dbname, systemTD.posX, systemTD.posY, systemTD.posZ
                )
                oldCount += 1
                if not (systemTD.posX == sysPosX and
                        systemTD.posY == sysPosY and
                        systemTD.posZ == sysPosZ):
                    tdenv.WARN("System '{}' has different coordinates:", sysName)
                    tdenv.WARN("   database: {}, {}, {}", systemTD.posX, systemTD.posY, systemTD.posZ)
                    tdenv.WARN("     netlog: {}, {}, {}", sysPosX, sysPosY, sysPosZ)
            else:
                # new system
                tdenv.NOTE("New system '{}' ({}, {}, {}, '{}')",
                    sysName, sysPosX, sysPosY, sysPosZ, utcDate
                )
                newCount += 1
                if not optShow:
                    tdb.addLocalSystem(
                        sysName.upper(),
                        sysPosX, sysPosY, sysPosZ,
                        added=self.ADDED_NAME,
                        modified=utcDate,
                        commit=False
                    )
                    addCount += 1

        tdenv.NOTE("Found {:>3} System(s) altogether.", sysCount)
        if oldCount:
            tdenv.NOTE("      {:>3} old", oldCount)
        if newCount:
            tdenv.NOTE("      {:>3} new", newCount)
        if addCount:
            tdenv.NOTE("      {:>3} added", addCount)
            tdb.getDB().commit()
            tdenv.NOTE("Export Table 'System'")
            _, path = csvexport.exportTableToFile(tdb, tdenv, "System")

    def getLogDirList(self):
        """
        get all netLog files
        """
        tdenv   = self.tdenv
        logDate = self.getOption("date")
        logLast = self.getOption("last")

        self.logDate = None
        if isinstance(logDate, str):
            self.fmtDate = len(logDate)
            fmt = self.DATE_FORMATS.get(self.fmtDate, (None, None, None))
            if fmt[0]:
                tdenv.DEBUG0("date format: {}", fmt[0])
                try:
                    self.logDate = datetime.strptime(logDate, fmt[0])
                except ValueError:
                    pass
            if self.logDate:
                globDat = self.logDate.strftime(fmt[2])
                self.logGlob = "netLog." + globDat + "*.log"
            else:
                raise PluginException(
                    "Wrong date '{}' format. Must be in the form of '{}'".format(
                    logDate,
                    "','".join([d[1] for d in self.DATE_FORMATS.values()]))
                )
            tdenv.NOTE("using date: {}", self.logDate.strftime(fmt[0]))
        tdenv.NOTE("using pattern: {}", self.logGlob)

        self.filePathList = []
        for filePath in sorted(self.logPath.glob(self.logGlob)):
            tdenv.DEBUG0("logfile: {}", str(filePath))
            self.filePathList.append(filePath)

        if logLast and len(self.filePathList) > 1:
            del self.filePathList[:-1]

        listLen = len(self.filePathList)
        if listLen == 0:
            raise PluginException("No logfile found.")
        elif listLen == 1:
            tdenv.NOTE("Found one logfile.")
        else:
            tdenv.NOTE("Found {} logfiles.", listLen)

    def run(self):
        tdb, tdenv = self.tdb, self.tdenv

        tdenv.DEBUG0("show: {}", self.getOption("show"))
        tdenv.DEBUG0("last: {}", self.getOption("last"))
        tdenv.DEBUG0("date: {}", self.getOption("date"))

        # Ensure the cache is built and reloaded.
        tdb.reloadCache()
        tdb.load(maxSystemLinkLy=tdenv.maxSystemLinkLy)

        self.getLogDirList()
        self.parseLogDirList()

        # We did all the work
        return False



import os
import json
import pathlib
from datetime import datetime, timezone

from .. import csvexport
from . import PluginException, ImportPluginBase


def snapToGrid32(val):
    try:
        val = float(val)
        corr = -0.5 if val < 0 else 0.5
        pos = int(val*32+corr)/32
    except:
        pos = None
        pass
    return pos

def getYNfromService(obj, key):
    return "Y" if key in obj else "N"

class JournalStation(object):
    __slots__ = (
        'lsFromStar', 'blackMarket', 'maxPadSize',
        'market', 'shipyard', 'outfitting',
        'rearm', 'refuel', 'repair',
        'planetary', 'modified'
    )

    def __init__(
        self, lsFromStar=0, blackMarket='?', maxPadSize='?',
        market='?', shipyard='?', outfitting='?',
        rearm='?', refuel='?', repair='?',
        planetary='?', modified='now'
    ):
        self.lsFromStar = lsFromStar
        self.blackMarket = blackMarket
        self.maxPadSize = maxPadSize
        self.market = market
        self.shipyard = shipyard
        self.outfitting = outfitting
        self.rearm = rearm
        self.refuel = refuel
        self.repair = repair
        self.planetary = planetary
        self.modified = modified

    def __str__(self):
        return "{}ls Pad:{} Mkt:{} Blk:{} Shp:{} Out:{} Arm:{} Ref:{} Rep:{} Plt:{}".format(
            self.lsFromStar, self.maxPadSize, self.market,
            self.blackMarket, self.shipyard, self.outfitting,
            self.rearm, self.refuel, self.repair, self.planetary
        )

class ImportPlugin(ImportPluginBase):
    """
    Plugin that parses the Journal file and add or update systems and stations.
    """

    logGlob = "Journal.*.log"
    ADDED_NAME = 'Journal'
    LOGDIR_NAME = "FDEVJRNDIR"
    DATE_FORMATS = {
         2: ("%y",       "YY",         "%y"),
         4: ("%Y",       "YYYY",       "%y"),
         5: ("%y-%m",    "YY-MM",      "%y%m"),
         7: ("%Y-%m",    "YYYY-MM",    "%y%m"),
         8: ("%y-%m-%d", "YY-MM-DD",   "%y%m%d"),
        10: ("%Y-%m-%d", "YYYY-MM-DD", "%y%m%d"),
    }
    ignoreSysNames = [
        'TRAINING',
        'DESTINATION',
    ]
    pluginOptions = {
        'show': "Only show the system or station. Don't update the DB.",
        'last': "Only parse the last (newest) Journal file.",
        'date': "Only parse Journal files from date, format=[YY]YY[-MM[-DD]].",
    }
    filePathList = []
    sysList = {}
    stnList = {}
    blkList = []

    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)

        logDirName = os.getenv(self.LOGDIR_NAME, None)
        if not logDirName:
            raise PluginException(
                "Environment variable '{}' not set "
                "(see 'README.md' for help)."
                .format(self.LOGDIR_NAME)
            )
        tdenv.NOTE("{}={}", self.LOGDIR_NAME, logDirName)

        self.logPath = pathlib.Path(logDirName)
        if not self.logPath.is_dir():
            raise PluginException(
                "{}: is not a directory.".format(
                    str(self.logPath)
                )
            )

    def getJournalDirList(self):
        """
        get all Journal files
        """
        tdenv = self.tdenv
        optDate = self.getOption("date")
        logLast = self.getOption("last")

        logDate = None
        if isinstance(optDate, str):
            fmtLen = len(optDate)
            fmtDate = self.DATE_FORMATS.get(fmtLen, None)
            if fmtDate:
                tdenv.DEBUG0("date format: {}", fmtDate[0])
                try:
                    logDate = datetime.strptime(optDate, fmtDate[0])
                except ValueError:
                    logDate = None
                    pass
            if logDate:
                globDat = logDate.strftime(fmtDate[2])
                self.logGlob = "Journal." + globDat + "*.log"
                tdenv.NOTE("using date: {}", logDate.strftime(fmtDate[0]))
            else:
                raise PluginException(
                    "Wrong date '{}' format. Must be in the form of '{}'"
                    .format(
                        optDate,
                        "','".join([d[1] for d in self.DATE_FORMATS.values()])
                    )
                )
        tdenv.NOTE("using pattern: {}", self.logGlob)

        for filePath in sorted(self.logPath.glob(self.logGlob)):
            tdenv.DEBUG0("logfile: {}", str(filePath))
            self.filePathList.append(filePath)

        listLen = len(self.filePathList)
        if listLen == 0:
            raise PluginException("No journal file found.")
        elif listLen == 1:
            tdenv.NOTE("Found one journal file.")
        else:
            tdenv.NOTE("Found {} journal files.", listLen)

        if logLast and listLen > 1:
            del self.filePathList[:-1]

    def parseJournalDirList(self):
        """
        parse Journal files
        see: https://forums.frontier.co.uk/showthread.php/275151-Commanders-log-manual-and-data-sample
        """
        tdenv = self.tdenv

        logSysList = {}
        stnSysList = {}
        blkStnList = []
        for filePath in self.filePathList:
            tdenv.NOTE("parsing '{}'", filePath.name)
            aktStation = False
            inMultiCrew = False
            sysCount = stnCount = blkCount = 0
            with filePath.open() as logFile:
                lineCount = 0
                statHeader = True
                for line in logFile:
                    lineCount += 1
                    try:
                        # parse the json-event-line of the journal
                        event = json.loads(line)
                        if inMultiCrew:
                            # ignore all events in multicrew except
                            if event["event"] == "QuitACrew":
                                inMultiCrew = False
                            else:
                                tdenv.DEBUG1("event '{}' ignored (multicrew)", event["event"])
                            continue
                        logDate = datetime.strptime(
                            event["timestamp"], "%Y-%m-%dT%H:%M:%SZ"
                        ).replace(tzinfo=timezone.utc)
                        if statHeader:
                            # check the first line
                            statHeader = False
                            if event["event"] == "Fileheader":
                                if "beta" in event["gameversion"].lower():
                                    # don't parse data from beta versions
                                    tdenv.NOTE("Ignoring Beta-Version.")
                                    break
                                # ignore the header-event
                                continue
                            else:
                                # don't stop parsing if it's not the header-line
                                tdenv.WARN("Doesn't seem do be a FDEV Journal file")
                        if event["event"] == "Location":
                            # convert event before the if-elif starts
                            if event.get("Docked", False):
                                event["event"] = "Docked"
                                tdenv.DEBUG0("   EVENT: Changed Location to Docked")
                        # if-elif starts here
                        if event["event"] == "FSDJump":
                            sysCount += 1
                            sysDate = logDate
                            sysName = event["StarSystem"]
                            sysPosA = event["StarPos"]
                            sysPosX, sysPosY, sysPosZ = sysPosA[0], sysPosA[1], sysPosA[2]
                            sysPosX = snapToGrid32(sysPosX)
                            sysPosY = snapToGrid32(sysPosY)
                            sysPosZ = snapToGrid32(sysPosZ)
                            tdenv.DEBUG0(
                                "  SYSTEM: {} {} {} {} {}",
                                sysDate, sysName, sysPosX, sysPosY, sysPosZ
                            )
                            logSysList[sysName] = (sysPosX, sysPosY, sysPosZ, sysDate)
                        elif event["event"] == "Docked":
                            stnCount += 1
                            sysName = event["StarSystem"]
                            stnList = stnSysList.get(sysName, None)
                            if not stnList:
                                stnList = stnSysList[sysName] = {}
                            stnDate = logDate
                            stnName = event["StationName"]
                            jrnStation = JournalStation(modified=stnDate)
                            jrnStation.lsFromStar = int(event.get("DistFromStarLS", 0) + 0.5)
                            stnType = event.get("StationType", None)
                            if stnType:
                                # conclusions from the stationtype
                                jrnStation.planetary = "Y" if stnType == "SurfaceStation" else "N"
                                jrnStation.maxPadSize = "M" if stnType.startswith("Outpost") else "L"
                            stnServices = event.get("StationServices", None)
                            if stnServices:
                                # station services since ED update 2.4
                                jrnStation.blackMarket = getYNfromService(stnServices, 'BlackMarket')
                                jrnStation.market = getYNfromService(stnServices, 'Commodities')
                                jrnStation.shipyard = getYNfromService(stnServices, 'Shipyard')
                                jrnStation.outfitting = getYNfromService(stnServices, 'Outfitting')
                                jrnStation.rearm = getYNfromService(stnServices, 'Rearm')
                                jrnStation.refuel = getYNfromService(stnServices, 'Refuel')
                                jrnStation.repair = getYNfromService(stnServices, 'Repair')
                            tdenv.DEBUG0(
                                " STATION: {} {}/{} {}",
                                stnDate, sysName, stnName, str(jrnStation)
                            )
                            stnList[stnName] = jrnStation
                            aktStation = True
                            sysPosA = event.get("StarPos", None)
                            if sysPosA:
                                # we got system data inside a docking event
                                # use it (changed Location or maybe EDDN capture)
                                sysCount += 1
                                sysDate = logDate
                                sysPosX, sysPosY, sysPosZ = sysPosA[0], sysPosA[1], sysPosA[2]
                                sysPosX = snapToGrid32(sysPosX)
                                sysPosY = snapToGrid32(sysPosY)
                                sysPosZ = snapToGrid32(sysPosZ)
                                tdenv.DEBUG0("  SYSTEM: {} {} {} {} {}", sysDate, sysName, sysPosX, sysPosY, sysPosZ)
                                logSysList[sysName] = (sysPosX, sysPosY, sysPosZ, sysDate)
                        elif event["event"] == "MarketSell" and aktStation:
                            # check for BlackMarket
                            if event.get("BlackMarket", False):
                                stnBlackMarket = (sysName, stnName)
                                if stnBlackMarket not in blkStnList:
                                    tdenv.DEBUG0("B/MARKET: {}/{}", sysName, stnName)
                                    blkCount += 1
                                    blkStnList.append(stnBlackMarket)
                        elif event["event"] == "Undocked":
                            aktStation = False
                        elif event["event"] == "JoinACrew":
                            inMultiCrew = True
                        else:
                            tdenv.DEBUG1("event '{}' ignored", event["event"])
                    except:
                        tdenv.WARN("Something wrong with line {} (skipped it).", lineCount)

            tdenv.NOTE(
                "Found {} System{}, {} Station{} and {} BlackMarket{}.",
                sysCount, "" if sysCount == 1 else "s",
                stnCount, "" if stnCount == 1 else "s",
                blkCount, "" if blkCount == 1 else "s",
            )
        self.sysList = logSysList
        self.stnList = stnSysList
        self.blkList = blkStnList

    def updateJournalSysList(self):
        """
        check the found systems and add them to the DB if new.
        """
        tdb, tdenv = self.tdb, self.tdenv
        optShow = self.getOption("show")

        if not optShow:
            try:
                idJournal = tdb.lookupAdded(self.ADDED_NAME)
            except KeyError:
                tdenv.WARN("Entry '{}' not found in 'Added' table.", self.ADDED_NAME)
                tdenv.WARN("Trying to add it myself.")
                db = tdb.getDB()
                cur = db.cursor()
                cur.execute(
                    "INSERT INTO Added(name) VALUES(?)",
                    [self.ADDED_NAME]
                )
                db.commit()
                tdenv.NOTE("Export Table 'Added'")
                _, path = csvexport.exportTableToFile(tdb, tdenv, "Added")
                pass

        addCount = oldCount = newCount = 0
        for sysName in sorted(self.sysList):
            sysPosX, sysPosY, sysPosZ, sysDate = self.sysList[sysName]
            utcDate = sysDate.astimezone(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
            tdenv.DEBUG0(
                "log system '{}' ({}, {}, {}, '{}')",
                sysName, sysPosX, sysPosY, sysPosZ, utcDate
            )
            if sysName.upper() in self.ignoreSysNames:
                tdenv.NOTE("Ignoring system: '{}'", sysName)
                continue
            systemTD = tdb.systemByName.get(sysName.upper(), None)
            if systemTD:
                # we allready know the system, check coords
                tdenv.DEBUG0(
                    "Old system '{}' ({}, {}, {})",
                    systemTD.dbname, systemTD.posX, systemTD.posY, systemTD.posZ
                )
                oldCount += 1
                if not (systemTD.posX == sysPosX and
                        systemTD.posY == sysPosY and
                        systemTD.posZ == sysPosZ):
                    tdenv.WARN("System '{}' has different coordinates:", sysName)
                    tdenv.WARN("   database: {}, {}, {}", systemTD.posX, systemTD.posY, systemTD.posZ)
                    tdenv.WARN("    Journal: {}, {}, {}", sysPosX, sysPosY, sysPosZ)
            else:
                # it's a new system
                newCount += 1
                if optShow:
                    # display only
                    tdenv.NOTE(
                        "New system '{}' ({}, {}, {}, '{}')",
                        sysName, sysPosX, sysPosY, sysPosZ, utcDate
                    )
                else:
                    # add it to the database
                    # the function will output something
                    tdb.addLocalSystem(
                        sysName.upper(),
                        sysPosX, sysPosY, sysPosZ,
                        added=self.ADDED_NAME,
                        modified=utcDate,
                        commit=False
                    )
                    addCount += 1

        # output statistics
        allCount = oldCount + newCount
        tdenv.NOTE(
            "Found {:>3} System{} altogether.",
            allCount, "" if allCount == 1 else "s",
        )
        for iCount, iText in [
            (oldCount, "old"), (newCount, "new"), (addCount, "added"),
        ]:
            tdenv.NOTE("      {:>3} {}", iCount, iText)
        if addCount:
            tdb.getDB().commit()
            tdenv.NOTE("Export Table 'System'")
            _, path = csvexport.exportTableToFile(tdb, tdenv, "System")

    def updateJournalStnList(self):
        """
        check the found stations and
        add them to the DB if new or
        update them in the DB if changed.
        """
        tdb, tdenv = self.tdb, self.tdenv
        optShow = self.getOption("show")

        addCount = oldCount = newCount = updCount = 0
        for sysName in sorted(self.stnList):
            if sysName.upper() in self.ignoreSysNames:
                tdenv.NOTE("Ignoring system: '{}'", sysName)
                continue
            system = tdb.systemByName.get(sysName.upper(), None)
            if not (system or optShow):
                # only warn if we are not in show mode
                # otherwise we could have addded the system before
                tdenv.WARN("System '{}' unknown.", sysName)
                continue
            for stnName in sorted(self.stnList[sysName]):
                jrnStation = self.stnList[sysName][stnName]
                utcDate = jrnStation.modified.astimezone(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
                station = None
                if system:
                    # system could be None in show mode and the lookup
                    # function would try very hard to find the station
                    try:
                        station = tdb.lookupStation(stnName, system)
                    except LookupError:
                        pass

                if (sysName, stnName) in self.blkList:
                    # BlackMarket found
                    jrnStation.blackMarket = "Y"
                elif station:
                    if jrnStation.blackMarket == "?":
                        # don't change current value if new one is unknown
                        jrnStation.blackMarket = station.blackMarket

                tdenv.DEBUG0(
                    "log station '{}/{}' ({}, '{}')",
                    sysName, stnName, str(jrnStation), utcDate
                )

                if not station:
                    # must be a new station
                    newCount += 1
                    if optShow:
                        # display only
                        tdenv.NOTE(
                            "New station '{}/{}' ({}, '{}')",
                            sysName, stnName, str(jrnStation), utcDate
                        )
                    else:
                        # add it to the database
                        # the function will output something
                        station = tdb.addLocalStation(
                            system=system,
                            name=stnName,
                            lsFromStar=jrnStation.lsFromStar,
                            blackMarket=jrnStation.blackMarket,
                            maxPadSize=jrnStation.maxPadSize,
                            market=jrnStation.market,
                            shipyard=jrnStation.shipyard,
                            outfitting=jrnStation.outfitting,
                            rearm=jrnStation.rearm,
                            refuel=jrnStation.refuel,
                            repair=jrnStation.repair,
                            planetary=jrnStation.planetary,
                            modified=utcDate,
                            commit=False,
                        )
                        addCount += 1
                else:
                    oldCount += 1
                    tdenv.DEBUG0(
                        "Old station '{}' ({}ls Pad:{} Mkt:{} Blk:{} Shp:{} Out:{} Arm:{} Ref:{} Rep:{} Plt:{})",
                        station.name(), station.lsFromStar, station.maxPadSize, station.market,
                        station.blackMarket, station.shipyard, station.outfitting,
                        station.rearm, station.refuel, station.repair, station.planetary
                    )
                    if not optShow:
                        if (station.lsFromStar != jrnStation.lsFromStar and
                            station.blackMarket == jrnStation.blackMarket and
                            station.maxPadSize == jrnStation.maxPadSize and
                            station.market == jrnStation.market and
                            station.shipyard == jrnStation.shipyard and
                            station.outfitting == jrnStation.outfitting and
                            station.rearm == jrnStation.rearm and
                            station.refuel == jrnStation.refuel and
                            station.repair == jrnStation.repair and
                            station.planetary == jrnStation.planetary
                        ):
                            # ignore 15% deviation if it's the only change
                            lsMin = int(station.lsFromStar * 0.85)
                            lsMax = int(station.lsFromStar*1.15 + 1)
                            if lsMin <= jrnStation.lsFromStar <= lsMax:
                                tdenv.DEBUG0(
                                    "ignore 15% deviation ({}ls ~ {}ls)",
                                    jrnStation.lsFromStar, station.lsFromStar
                                )
                                jrnStation.lsFromStar = station.lsFromStar
                        # the function will do it's own check and output
                        # something if the station is updated
                        if tdb.updateLocalStation(
                            station=station,
                            lsFromStar=jrnStation.lsFromStar,
                            blackMarket=jrnStation.blackMarket,
                            maxPadSize=jrnStation.maxPadSize,
                            market=jrnStation.market,
                            shipyard=jrnStation.shipyard,
                            outfitting=jrnStation.outfitting,
                            rearm=jrnStation.rearm,
                            refuel=jrnStation.refuel,
                            repair=jrnStation.repair,
                            planetary=jrnStation.planetary,
                            modified=utcDate,
                            commit=False,
                        ):
                            updCount += 1

        # output statistics
        allCount = oldCount + newCount
        tdenv.NOTE(
            "Found {:>3} Station{} altogether.",
            allCount, "" if allCount == 1 else "s",
        )
        for iCount, iText in [
            (oldCount, "old"), (updCount, "updated"),
            (newCount, "new"), (addCount, "added"),
        ]:
            tdenv.NOTE("      {:>3} {}", iCount, iText)
        if (updCount+addCount) > 0:
            tdb.getDB().commit()
            tdenv.NOTE("Export Table 'Station'")
            _, path = csvexport.exportTableToFile(tdb, tdenv, "Station")

    def run(self):
        tdb, tdenv = self.tdb, self.tdenv

        tdenv.DEBUG0("show: {}", self.getOption("show"))
        tdenv.DEBUG0("last: {}", self.getOption("last"))
        tdenv.DEBUG0("date: {}", self.getOption("date"))

        # Ensure the cache is built and reloaded.
        tdb.reloadCache()
        tdb.load(maxSystemLinkLy=tdenv.maxSystemLinkLy)

        self.getJournalDirList()
        self.parseJournalDirList()
        self.updateJournalSysList()
        self.updateJournalStnList()

        # We did all the work
        return False
